use rand::prelude::*;
use rand_distr::{Distribution, Normal};
use serde::{Deserialize, Serialize};
use std::fs::File;
use std::io::BufWriter;
use std::io::{Read, Write};

#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct NeuralNet {
	matrix: Vec<Vec<f64>>,
	biases: Vec<Vec<f64>>,
	// since each layer has a different # of neurons,
	// some values in this 2D array are never read
	weights: Vec<Vec<f64>>,
}

impl NeuralNet {
	pub fn new_random_weights(
		num_inputs: usize,
		num_outputs: usize,
		num_hidden_layers: usize,
		num_nodes_per_hidden_layer: usize,
	) -> Self {
		let mut rng = rand::thread_rng();

		let mut matrix = Vec::new();
		let mut biases = Vec::new();

		matrix.push(vec![0.0; num_inputs]); // Input layer
		biases.push(rand_vec(&mut rng, num_inputs));

		for _ in 0..num_hidden_layers {
			matrix.push(vec![0.0; num_nodes_per_hidden_layer]);
			biases.push(rand_vec(&mut rng, num_nodes_per_hidden_layer));
		}

		matrix.push(vec![0.0; num_outputs]);
		biases.push(rand_vec(&mut rng, num_outputs));

		let side = num_inputs.max(num_outputs).max(num_nodes_per_hidden_layer);
		let mut weights = Vec::new();
		for _ in 0..side {
			let mut weight_col = Vec::new();
			for _ in 0..side {
				weight_col.push(rng.gen_range(-100.0..100.0));
			}
			weights.push(weight_col);
		}

		Self {
			matrix,
			weights,
			biases,
		}
	}

	pub fn compute(&mut self, input: Vec<f64>) -> &[f64] {
		if input.len() != self.matrix[0].len() {
			panic!("Input array was not of correct length");
		}

		// NOTE - we're not biasing the input. It's probably fine
		self.matrix = self.biases.clone();
		self.matrix[0] = input;

		for new_i in 1..self.matrix.len() {
			let old_i = new_i - 1;
			for new_j in 0..self.matrix[new_i].len() {
				for old_j in 0..self.matrix[old_i].len() {
					self.matrix[new_i][new_j] +=
						self.matrix[old_i][old_j] * self.weights[old_j][new_j]
				}
				self.matrix[new_i][new_j] = sigmoid(self.matrix[new_i][new_j]);
			}
		}

		self.matrix.last().unwrap()
	}

	pub fn reproduce(&self, other: &Self) -> Self {
		let mut rng = rand::thread_rng();
		let crossover_point = rng.gen_range(0..self.weights.len());

		let mut i = 0;
		let weights = self
			.weights
			.iter()
			.zip(&other.weights)
			.map(|(row0, row1)| {
				row0.iter()
					.zip(row1)
					.map(|(x1, x2)| {
						i += 1;
						merge_values(&mut rng, *x1, *x2, i < crossover_point)
					})
					.collect()
			})
			.collect();

		i = 0;
		let biases = self
			.biases
			.iter()
			.zip(&other.biases)
			.map(|(row0, row1)| {
				row0.iter()
					.zip(row1)
					.map(|(x1, x2)| {
						i += 1;
						merge_values(&mut rng, *x1, *x2, i < crossover_point)
					})
					.collect()
			})
			.collect();

		Self {
			matrix: self.matrix.clone(),
			biases,
			weights,
		}
	}

	pub fn save(&self, filename: &str) {
		let mut buffer = BufWriter::new(File::create(filename).unwrap());

		buffer
			.write_all(&bincode::serialize(&self).unwrap())
			.unwrap();
		buffer.flush().unwrap();
	}

	pub fn load(filename: &str) -> Self {
		let mut buffer = Vec::new();
		File::open(filename)
			.unwrap()
			.read_to_end(&mut buffer)
			.unwrap();
		bincode::deserialize(&buffer).unwrap()
	}
}

// -1.0 to 1.0
fn sigmoid(x: f64) -> f64 {
	// Handle infinities and -infinities
	if x > 100.0 {
		return 1.0;
	}

	if x < -100.0 {
		return -1.0;
	}

	let e_x = x.exp();

	2.0 * e_x / (e_x + 1.0) - 1.0
}

fn rand_vec(rng: &mut ThreadRng, len: usize) -> Vec<f64> {
	(0..len).map(|_| rng.gen()).collect()
}

fn merge_values(rng: &mut ThreadRng, p1: f64, p2: f64, cond: bool) -> f64 {
	// TODO should be lazy static
	let normal = Normal::new(0.0, 1.0).unwrap();
	let mutation_chance = rng.gen::<f64>() > 0.50;

	let mutation_term = if mutation_chance {
		normal.sample(rng)
	} else {
		0.0
	};
	let merged_term = if cond { p1 } else { p2 };

	mutation_term + merged_term
}
