use crate::neural_net::NeuralNet;
use crate::planet::Planet;
use kiss3d::scene::SceneNode;
use kiss3d::window::Window;
use na::Translation;
use na::{Point, Vector3};

const THREE_OVER_FOUR_PI: f32 = 3.0 / (4.0 * std::f32::consts::PI);

pub struct Probe {
	pub pos: Vector3<f32>,
	pub vel: Vector3<f32>,
	pub mass: f32,
	node: Option<SceneNode>,
	pub brain: NeuralNet,
}

impl Probe {
	pub fn new(
		window: Option<&mut Window>,
		pos: Vector3<f32>,
		vel: Vector3<f32>,
		mass: f32,
	) -> Self {
		let node = window.map(|w| {
			let mut n = w.add_sphere(1.0);
			n.set_color(1.0, 0.0, 0.0);
			n
		});

		let mut s = Self {
			pos,
			vel,
			mass,
			node,
			brain: NeuralNet::new_random_weights(8, 4, 1, 10),
		};

		let r = s.radius();
		if let Some(n) = &mut s.node {
			n.set_local_scale(r, r, r);
		}

		s
	}

	pub fn new_with_brain(
		brain: NeuralNet,
		pos: Vector3<f32>,
		vel: Vector3<f32>,
		mass: f32,
	) -> Self {
		Self {
			pos,
			vel,
			mass,
			node: None,
			brain,
		}
	}

	pub fn new_with_window_and_brain(
		window: &mut Window,
		brain: NeuralNet,
		pos: Vector3<f32>,
		vel: Vector3<f32>,
		mass: f32,
	) -> Self {
		let mut node = window.add_sphere(1.0);
		node.set_color(1.0, 0.0, 0.0);

		let mut s = Self {
			pos,
			vel,
			mass,
			node: Some(node),
			brain,
		};

		let r = s.radius();
		if let Some(n) = &mut s.node {
			n.set_local_scale(r, r, r);
		}

		s
	}

	pub fn step(&mut self, delta: f32) {
		self.pos += self.vel * delta;
		if let Some(n) = &mut self.node {
			n.set_local_translation(Translation::from(self.pos));
		}
	}

	pub fn collides_with(&self, planet: &Planet) -> bool {
		let dist_squared = na::distance_squared(&Point::from(self.pos), &Point::from(planet.pos));
		let min_dist = self.radius() + planet.radius;

		dist_squared < min_dist * min_dist
	}

	pub fn calculate_gravity(&mut self, planet: &Planet, delta: f32) {
		const G: f32 = 0.0001;
		let diff = planet.pos - self.pos;
		let dist_squared = diff.magnitude_squared();

		let force_magnitude = delta * G * planet.mass / dist_squared;
		self.vel += force_magnitude * diff.normalize();
	}

	pub fn radius(&self) -> f32 {
		(THREE_OVER_FOUR_PI * self.mass).cbrt()
	}

	pub fn think(&mut self, planets: &[Planet], time: f32, delta: f32) {
		let closest_planet = planets.iter().fold(
			(std::f32::MAX, Vector3::new(0.0, 0.0, 0.0), 0.0),
			|(best_dist_squared, best_vec, best_radius), cur| {
				let new_dist_squared = (self.pos - cur.pos).norm_squared();
				if new_dist_squared < best_dist_squared {
					(new_dist_squared, cur.pos, cur.radius)
				} else {
					(best_dist_squared, best_vec, best_radius)
				}
			},
		);

		let radii = closest_planet.2 + self.radius();
		if radii * radii > closest_planet.0 {
			// Collision
			return;
		}

		let v = self.pos - closest_planet.1;

		// Mass, distance to closest planet, velocity
		let output = self.brain.compute(vec![
			self.mass as f64,
			v.x as f64,
			v.y as f64,
			v.z as f64,
			self.vel.x as f64,
			self.vel.y as f64,
			self.vel.z as f64,
			time as f64,
		]);

		let x = output[0] as f32;
		let y = output[1] as f32;
		let z = output[2] as f32;
		let scale = output[3] as f32 * 0.1 * delta;

		if scale < 0.0 {
			return; // Do nothing
		}

		let delta_v = Vector3::new(x, y, z).normalize() * scale;

		self.mass -= scale;

		self.vel += delta_v;
	}

	pub fn consume(&mut self, planet: &mut Planet, delta: f32) {
		let dist_squared = na::distance_squared(&Point::from(self.pos), &Point::from(planet.pos));
		let min_dist = self.radius() + planet.radius + 10.0; // Can eat at distance

		if dist_squared > min_dist * min_dist {
			return;
		}

		let amount = delta.min(planet.mass) * 0.001;
		self.mass += amount;
		planet.mass -= amount;
	}
}
