use crate::neural_net::NeuralNet;
use crate::planet::Planet;
use crate::probe::Probe;
use kiss3d::window::Window;
use na::Vector3;
use rand::prelude::*;

const MAX_ABS_COORD: f32 = 1_000_000.0;
const MIN_VAL: f32 = 20.0;
const MAX_VAL: f32 = 20.0;
const MIN_SPD: f32 = 0.05;
const MAX_SPD: f32 = 0.05;

pub struct Universe {
	pub planets: Vec<Planet>,
	pub probes: Vec<Probe>,
	pub brain: NeuralNet,
	fitness: f32,
	pub time: f32,
}

impl Universe {
	pub fn new(window: Option<&mut Window>, num_bodies: usize) -> Self {
		let mut planets = Vec::with_capacity(num_bodies);
		let mut rng = rand::thread_rng();
		planets.push(Planet::new(100_000.0, Vector3::new(0.0, 0.0, 0.0)));
		for _ in 1..num_bodies {
			let x = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let y = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let z = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);

			planets.push(Planet::new(100_000.0, Vector3::new(x, y, z)));
		}

		let probes = vec![Probe::new(
			window,
			Vector3::new(
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
			),
			Vector3::new(
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
			),
			1.0,
		)];

		let brain = probes[0].brain.clone();

		Self {
			planets,
			probes,
			brain,
			fitness: 0.0,
			time: 0.0,
		}
	}

	pub fn new_with_brain(brain: NeuralNet, num_bodies: usize) -> Self {
		let mut planets = Vec::with_capacity(num_bodies);
		let mut rng = rand::thread_rng();
		planets.push(Planet::new(100_000.0, Vector3::new(0.0, 0.0, 0.0)));
		for _ in 1..num_bodies {
			let x = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let y = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let z = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);

			planets.push(Planet::new(100_000.0, Vector3::new(x, y, z)));
		}

		let probes = vec![Probe::new_with_brain(
			brain.clone(),
			Vector3::new(
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
			),
			Vector3::new(
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
			),
			1.0,
		)];

		Self {
			planets,
			probes,
			brain,
			fitness: 0.0,
			time: 0.0,
		}
	}

	pub fn new_with_window_and_brain(
		window: &mut Window,
		brain: NeuralNet,
		num_bodies: usize,
	) -> Self {
		let mut planets = Vec::with_capacity(num_bodies);
		let mut rng = rand::thread_rng();
		planets.push(Planet::new(100_000.0, Vector3::new(0.0, 0.0, 0.0)));
		for _ in 1..num_bodies {
			let x = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let y = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);
			let z = rng.gen_range(-MAX_ABS_COORD..MAX_ABS_COORD);

			planets.push(Planet::new(100_000.0, Vector3::new(x, y, z)));
		}

		let probes = vec![Probe::new_with_window_and_brain(
			window,
			brain.clone(),
			Vector3::new(
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
				rng.gen_range(MIN_VAL..=MAX_VAL),
			),
			Vector3::new(
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
				rng.gen_range(MIN_SPD..=MAX_SPD),
			),
			1.0,
		)];

		Self {
			planets,
			probes,
			brain,
			fitness: 0.0,
			time: 0.0,
		}
	}

	pub fn step(&mut self, delta: f32) -> bool {
		for p in &mut self.probes {
			p.think(&self.planets, self.time, delta);

			for _ in 0..100 {
				for planet in &mut self.planets {
					p.calculate_gravity(planet, delta / 100.0);

					if p.collides_with(planet) {
						p.vel *= 0.0;
					}

					p.consume(planet, delta);
				}

				p.step(delta / 100.0);

				// as close to 100 altitude for as long as possible
				//let v = (p.pos - self.planets[0].pos).norm() - 100.0;
				//self.fitness += 100.0 - v * v * 10.0
			}
		}

		for i in 0..self.probes.len() {
			if self.probes[i].mass < 0.0001 {
				self.probes.remove(i);
			}
		}

		for i in 0..self.planets.len() {
			if self.planets[i].mass < 0.0001 {
				self.planets.remove(i);
			}
		}

		self.time += delta;

		!self.probes.is_empty()
	}

	pub fn fitness(&self) -> f32 {
		if self.probes.is_empty() {
			return -100.0;
		}

		// self.fitness

		// as much energy as possible
		// self.probes[0].mass

		// get as high as possible
		self.probes[0].pos.norm()
	}
}
